/*
//  File mdlUtils.c
//
//  September 3 1992 - January 1994
//
//  Frederic Boulanger - boulange@supelec.fr
//  Ecole Superieure d Electricite - Service Informatique
//  Plateau de Moulon - 91192 Gif sur Yvette cedex
//  France
//
//  This file is part of the Module Description Language Compiler source set
//  Copyright Supelec, 1992-1994.
//  All Rights Reserved.
//  
// Redistribution and use are permitted provided that:
//   - distributions including source files retain this entire copyright 
//     notice and comment,
//   - distributions including binaries display the following acknowledgment: 
//     "This product includes software developed by Supelec and its contributors" 
//     in the documentation or other materials provided with the distribution and 
//     in all advertising material mentioning features or use of this software. 
// 
//  
//   THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED 
//   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF 
//   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
//
//  Author: Frederic Boulanger (Frederic.Boulanger@supelec.fr)
//          Ecole Superieure d Electricite - Service Informatique
//          Plateau de Moulon - 91192 Gif sur Yvette cedex
//          France
//
*/

/* This file contains miscellaneous functions for accessing      */
/* information.                                                  */
/*****************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "env.h"
#include "mdlErrors.h"
#include "mdlUtils.h"
#include "mdl.h"
#include "mdlTables.h"

static char dateStr[256];  /* Buffer for date string manipulations */

/***********************************************************************/
/* Function dateAndTime returns a pointer to a string representing the */
/* current date and time in the form :                                 */
/*    "DayOfWeek Month DayOfMonth Year at Hour:Minutes:Seconds"        */
/***********************************************************************/
char *dateAndTime() {
  time_t theClock;
  
  theClock = time(&theClock);
  strftime(dateStr,256,"%A %B %d %Y at %H:%M:%S", localtime(&theClock));
           /* 'man strftime' will teach you a lot about this function! */
  return dateStr;
}

/**************************************************************************/
/* Function fileName returns a pointer to the file name part of a full    */
/* file name. It looks at the last occurrence of the directory            */
/* separator character (dirSep, defined in env.h) and returns a pointer   */
/* to the next character in the string if it finds it. If not, it returns */
/* the full file name.                                                    */
/**************************************************************************/
char *fileName(char *fullFileName) {
  char *tmp;
  
  tmp = strrchr(fullFileName, dirSep);
  if (tmp == NULL)      /* If there is no directory separator in file name */
    tmp = fullFileName; /* return the file name */
  else                  /* If there is one */
    tmp++;              /* the file name begins at the next character. */
  return tmp;
}

/***********************************************************************/
/* Function checkAlloc allocates memory for nItems of size itemSize.   */
/* It checks for success of the operation and exits through fatalError */
/* if the allocation fails.                                            */
/***********************************************************************/
char *checkAlloc(size_t nItems, size_t itemSize) {
  char *storage;

  if (nItems > 0) {
    storage = calloc(nItems,itemSize);
    if (storage == 0) fatalError("Out of memory");
    return storage;
  }else
    return 0;
}

/***********************************************************************/
/* Function NameOfType yields the name of a type.                      */
/***********************************************************************/
char *NameOfType(atype* theType) {
  char buf[256];
  char *retval;
  typeList *params;
  
  strcpy(buf, theType->name);
  params = theType->params;
  if (params != 0)
    strcat(buf,"<");
  while (params != 0) {
    retval = NameOfType(params->theType);
    strcat(buf, retval);
    free(retval);
    strcat(buf,",");
    params = params->next;
  }
  if (theType->params != 0)
    buf[strlen(buf) - 1] = '>';
  retval = checkAlloc(strlen(buf) + 1, sizeof(char));
  strcpy(retval, buf);
  return retval;
}

/***************************************************************************/
/* Function cmpRawTypes compares two types without resolving typedefs and  */
/* returns 0 if they are equal.                                            */
/***************************************************************************/

int cmpRawTypes(atype *typ1, atype *typ2) {
  typeList *parm1, *parm2;
    
  if (strcmp(typ1->name, typ2->name) != 0)  /* If not same base name -> not same */
    return 1;
  parm1 = typ1->params;
  parm2 = typ2->params;
  while ((parm1 != 0)&&(parm2 != 0)) {      /* While params match -> Ok */
    if (cmpRawTypes(parm1->theType, parm2->theType) != 0)
      return 1;
    parm1 = parm1->next;
    parm2 = parm2->next;
  }
  if (parm1 != parm2)              /* If not same number of params */
    return 1;                      /* not same */
  else
    return 0;                      /* Same base name, same params -> same */
}

/***************************************************************************/
/* Function compareTypes compares two types and return 0 if they are equal.*/
/***************************************************************************/

int compareTypes(atype *typ1, atype *typ2) {
  typeList *parm1, *parm2;
    
  typ1 = resolveTypdef(typ1);      /* Resolve possible typedefs */
  typ2 = resolveTypdef(typ2);

  if (strcmp(typ1->name, typ2->name) != 0)  /* If not same base name -> not same */
    return 1;
  parm1 = typ1->params;
  parm2 = typ2->params;
  while ((parm1 != 0)&&(parm2 != 0)) {      /* While params match -> Ok */
    if (compareTypes(parm1->theType, parm2->theType) != 0)
      return 1;
    parm1 = parm1->next;
    parm2 = parm2->next;
  }
  if (parm1 != parm2)              /* If not same number of params */
    return 1;                      /* not same */
  else
    return 0;                      /* Same base name, same params -> same */
}

/***************************************************************************/
/* Function compareTypeNames compares two type names and return 0 if they  */
/* correspond to the same type (does not work with template types).        */
/***************************************************************************/

int compareTypeNames(char *typ1, char *typ2) {
  typeList *parm1, *parm2;
  
  atype t1, t2;
  atype *pt1, *pt2;
  
  t1.name = typ1;
  t1.params = 0;
  t2.name = typ2;
  t2.params = 0;
  
  pt1 = resolveTypdef(&t1);      /* Resolve possible typedefs */
  pt2 = resolveTypdef(&t2);

  if (strcmp(pt1->name, pt2->name) != 0)  /* If not same base name -> not same */
    return 1;
  parm1 = pt1->params;
  parm2 = pt2->params;
  while ((parm1 != 0)&&(parm2 != 0)) {      /* While params match -> Ok */
    if (compareTypes(parm1->theType, parm2->theType) != 0)
      return 1;
    parm1 = parm1->next;
    parm2 = parm2->next;
  }
  if (parm1 != parm2)              /* If not same number of params */
    return 1;                      /* not same */
  else
    return 0;                      /* Same base name, same params -> same */
}

/***************************************************************************/
/* Function compatTypes returns 0 if two types are compatible (can be made */
/* the same through identification of their parameters).                   */
/***************************************************************************/

int compatTypes(atype *typ1, atype *typ2) {
  typeList *parm1, *parm2;
  
  typ1 = resolveTypdef(typ1);      /* Resolve possible typedefs */
  typ2 = resolveTypdef(typ2);

  if (strcmp(typ1->name, typ2->name) != 0)  /* If not same base name -> not same */
    return 1;
  parm1 = typ1->params;
  parm2 = typ2->params;
  while ((parm1 != 0)&&(parm2 != 0)) {      /* While nb params match -> Ok */
    parm1 = parm1->next;
    parm2 = parm2->next;
  }
  if (parm1 != parm2)              /* If not same number of params */
    return 1;                      /* not compatible */
  else
    return 0;                      /* Same base name, same number of params -> compat */
}

/****************************************************************************/
/* Function resolveTemplate finds the real name of a type in the context of */
/* another type. The context type and its definition are assumed compatible.*/
/****************************************************************************/

char* resolveTemplate(atype *typ, atype *context, atype *contdef) {
  typeList *def, *cont;
  
  context = resolveTypdef(context); /* If typedef, resolve it */

  if (context->params == 0)  /* If not a template */
    return typ->name;
  def = contdef->params;     /* Look for name in context definition */
  cont = context->params;
  while (def != 0) {
    if (strcmp(typ->name, def->theType->name) == 0)
      return NameOfType(cont->theType);
    def = def->next;
    cont = cont->next;
  }
  return typ->name;          /* If not found, name is unchanged */
}

/****************************************************************************/
/* Function findTypdef returns 0 if there is no type definition for a given */
/* type, else, it returns a pointer on the equivalent type.                 */
/****************************************************************************/

atype* findTypdef(atype *typ) {
  typdefList *curTypdef;
  
  if (typ->params != 0)    /* If it is not a simple type, there is no typedef */
    return 0;
  curTypdef = theTypdefs;
  while(curTypdef != 0) {
    if (strcmp(typ->name, curTypdef->theDef->newType->name) == 0)
      return (curTypdef->theDef->oldType);
    curTypdef = curTypdef->next;
  }
  return 0;
}

/****************************************************************************/
/* Function resolveTypdef returns a pointer on the terminal equivalent type.*/
/****************************************************************************/

atype* resolveTypdef(atype *typ) {
  atype *tdef;
  
  while ((tdef = findTypdef(typ)) != 0)
    typ = tdef;
  return typ;
}

/*****************************************************************************/
/* Function printTemplate prints a C++ template declaration from a type list.*/
/*****************************************************************************/
void printTemplate(FILE* f, typeList *params) {
  int idx;
  char *t;

  if (params != 0) {
    fprintf(f,"template<");
    idx = 0;
    while(params != 0) {
      if (idx != 0)
        fprintf(f,", ");
      idx++;
	  t = NameOfType(params->theType);
      fprintf(f,"class %s",t);
	  free(t);
      params = params->next;
    }
    fprintf(f, ">\n");
  }
}
